# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-09
# @FilePath     : /LuWu/install_script.py
# @Desc         : Luwu一键安装脚本
import os
import platform
import subprocess
import sys


def stage_tip(num, msg, logger=None):
    total = 7
    message = "Step {}/{}: {}".format(num, total, msg)
    if logger is None:
        print(message)
    else:
        logger.info(message)


def run_cmd(cmd):
    proc = subprocess.Popen(
        cmd,
        stdin=subprocess.PIPE,
        stderr=sys.stderr,
        close_fds=True,
        stdout=sys.stdout,
        universal_newlines=True,
        shell=True,
        bufsize=1,
    )

    proc.communicate()
    code = proc.returncode
    if code != 0:
        exit(code)


def run_install():
    stage_tip(1, "检查系统类型是否支持...")
    system = platform.system()
    if system == "Darwin":
        print("System: MacOS")
        protoc_filename = "protoc-3.15.3-osx-x86_64.zip"
    elif system == "Linux":
        print("System: Linux")
        protoc_filename = "protoc-3.15.3-linux-x86_64.zip"
    elif system == "Windows":
        print("抱歉，LuWu 暂未支持在 Windows 上运行！")
        return -1
    else:
        print("不支持的系统类型 {} ！".format(system))
        return -1

    python_execute_path = sys.executable
    pip_execute_path = f"{python_execute_path} -m pip"

    stage_tip(2, "正在安装 requirements.txt ...")
    cmd = "{} install -r requirements.txt -i https://mirrors.aliyun.com/pypi/simple/ ".format(
        pip_execute_path
    )
    run_cmd(cmd)

    from loguru import logger

    stage_tip(3, "正在下载 TensorFlow Object Detection API ...", logger)
    cmd = "mkdir -p addons && cd addons && \
        rm -rf master* && \
        rm -rf models* && \
        wget https://codeload.github.com/tensorflow/models/zip/master &&\
        unzip master && mv models-master models && rm -rf master*"
    run_cmd(cmd)

    stage_tip(4, "正在下载并安装 protoc ...", logger)
    cmd = "cd addons && \
        mkdir -p protoc && \
        cd protoc && \
        rm -rf * && \
        rm -rf {protoc_filename}* && \
        wget https://github.com/protocolbuffers/protobuf/releases/download/v3.15.3/{protoc_filename} && \
        unzip {protoc_filename} && \
        rm -rf {protoc_filename}* && \
        export PATH=$PATH:`pwd`/bin && \
        cd ../models/research/ && \
        protoc object_detection/protos/*.proto --python_out=.".format(
        protoc_filename=protoc_filename
    )
    run_cmd(cmd)

    stage_tip(5, "正在安装 cocoapi ...", logger)
    cmd = "{} install git+https://github.com/philferriere/cocoapi.git#subdirectory=PythonAPI".format(
        pip_execute_path
    )
    run_cmd(cmd)

    stage_tip(6, "正在安装 TensorFlow Object Detection API ...", logger)
    cmd = "cd addons/models/research/ && \
        cp object_detection/packages/tf2/setup.py . && \
        {} -m pip install .".format(
        python_execute_path
    )
    run_cmd(cmd)

    stage_tip(7, "测试是否安装成功 ...", logger)
    cmd = "cd addons/models/research/ && \
        {} object_detection/builders/model_builder_tf2_test.py".format(
        python_execute_path
    )
    run_cmd(cmd)


if __name__ == "__main__":
    run_install()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-21
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /app/luwu/core/preprocess/data/templates/ImageClassifierDataGnenrator.py
# @Desc         :
import tensorflow as tf


def read_image(image_file_path: str):
    """从指定路径读取一张图片，并进行预处理

    Args:
        image_file_path (str): 待处理图片
    """
    image = tf.io.read_file(image_file_path)
    image = tf.io.decode_jpeg(image, channels=3)
    # imagenet数据集均值
    image_mean = [0.485, 0.456, 0.406]
    # imagenet数据集标准差
    image_std = [0.299, 0.224, 0.225]
    # 缩放图片
    image = tf.image.resize(image, [224, 224])
    # 将图片的像素值缩放到[0,1]之间
    image = tf.cast(image, dtype=tf.float32) / 255.0
    # 归一化
    image = (image - image_mean) / image_std
    return image



# -*- coding: utf-8 -*-
# @Date         : 2021-01-07
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /app/luwu/core/models/classifier/kerastuner/__init__.py
# @Desc         :
"""
这是还在规划中的功能，里面的代码目前都是无用的。
TODO:增加KerasTuner相关功能
"""


# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-07
# @FilePath     : /LuWu/luwu/core/models/complex/od/templates/project/eval.py
# @Desc         :
import time
import warnings

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from object_detection.utils import label_map_util
from object_detection.utils import visualization_utils as viz_utils
from PIL import Image

warnings.filterwarnings("ignore")  # Suppress Matplotlib warnings

PATH_TO_SAVED_MODEL = "my_model/saved_model"
PATH_TO_LABEL_MAP = "label_map.pbtxt"
PATH_TO_EVAL_IMAGES = [
    "test.image",
]


# 加载模型到内存
print("Loading model...", end="")
start_time = time.time()
detect_fn = tf.saved_model.load(PATH_TO_SAVED_MODEL)
end_time = time.time()
elapsed_time = end_time - start_time
print("Done! Took {} seconds".format(elapsed_time))

# 加载 label_map.pbtxt
category_index = label_map_util.create_category_index_from_labelmap(
    PATH_TO_LABEL_MAP, use_display_name=True
)

# 调用模型对指定图片进行目标检测
def load_image_into_numpy_array(path):
    """Load an image from file into a numpy array.

    Puts image into numpy array to feed into tensorflow graph.
    Note that by convention we put it into a numpy array with shape
    (height, width, channels), where channels=3 for RGB.

    Args:
      path: the file path to the image

    Returns:
      uint8 numpy array with shape (img_height, img_width, 3)
    """
    return np.array(Image.open(path))


for image_path in PATH_TO_EVAL_IMAGES:
    print("Running inference for {}... ".format(image_path), end="")
    image_np = load_image_into_numpy_array(image_path)
    # The input needs to be a tensor, convert it using `tf.convert_to_tensor`.
    input_tensor = tf.convert_to_tensor(image_np)
    # The model expects a batch of images, so add an axis with `tf.newaxis`.
    input_tensor = input_tensor[tf.newaxis, ...]
    # input_tensor = np.expand_dims(image_np, 0)
    detections = detect_fn(input_tensor)
    # All outputs are batches tensors.
    # Convert to numpy arrays, and take index [0] to remove the batch dimension.
    # We're only interested in the first num_detections.
    num_detections = int(detections.pop("num_detections"))
    detections = {
        key: value[0, :num_detections].numpy() for key, value in detections.items()
    }
    # detections是一个dict,结构如下：
    # - num_detections: a tf.int tensor with only one value, the number of detections [N].
    # - detection_boxes: a tf.float32 tensor of shape [N, 4] containing bounding box coordinates in the following order: [ymin, xmin, ymax, xmax].
    # - detection_classes: a tf.int tensor of shape [N] containing detection class index from the label file.
    # - detection_scores: a tf.float32 tensor of shape [N] containing detection scores.
    # - raw_detection_boxes: a tf.float32 tensor of shape [1, M, 4] containing decoded detection boxes without Non-Max suppression. M is the number of raw detections.
    # - raw_detection_scores: a tf.float32 tensor of shape [1, M, 90] and contains class score logits for raw detection boxes. M is the number of raw detections.
    # - detection_anchor_indices: a tf.float32 tensor of shape [N] and contains the anchor indices of the detections after NMS.
    # - detection_multiclass_scores: a tf.float32 tensor of shape [1, N, 90] and contains class score distribution (including background) for detection boxes in the image including background class.
    detections["num_detections"] = num_detections

    # detection_classes should be ints.
    detections["detection_classes"] = detections["detection_classes"].astype(np.int64)

    image_np_with_detections = image_np.copy()

    viz_utils.visualize_boxes_and_labels_on_image_array(
        image_np_with_detections,
        detections["detection_boxes"],
        detections["detection_classes"],
        detections["detection_scores"],
        category_index,
        use_normalized_coordinates=True,
        max_boxes_to_draw=200,
        min_score_thresh=0.30,
        agnostic_mode=False,
    )

    plt.figure()
    plt.imshow(image_np_with_detections)
    print("Done")
plt.show()



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-09
# @FilePath     : /LuWu/luwu/utils/cmd_util.py
# @Desc         :
import sys
import os
import subprocess
import shlex


def run_cmd(cmd, raise_exception=True):
    """使用shell执行指定命令，并将执行结果实时输出，返回命令退出的状态码

    Args:
        cmd (str): 待执行的shell命令

    Returns:
        int: shell退出时返回的状态码
    """
    if "JPY_PARENT_PID" in os.environ:
        # code, output = subprocess.getstatusoutput(cmd)
        p = subprocess.Popen(
            shlex.split(cmd), stdout=subprocess.PIPE, stderr=subprocess.STDOUT
        )
        while True:
            output = p.stdout.read1(1024).decode("utf-8")
            print(output, end="")
            if p.poll() is not None:
                break
        code = p.returncode
    else:
        proc = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stderr=sys.stderr,
            close_fds=True,
            stdout=sys.stdout,
            universal_newlines=True,
            shell=True,
            bufsize=1,
        )
        proc.communicate()
        code = proc.returncode
    if raise_exception and code != 0:
        raise Exception("命令运行过程中出错！")
    return code


def get_python_execute_path():
    """获取luwu安装到的python名称

    Returns:
        [type]: [description]
    """
    python_execute_path = sys.executable
    return python_execute_path


# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-09
# @FilePath     : /LuWu/luwu/core/models/complex/od/models.py
# @Desc         :
import os
from glob import glob

import tensorflow as tf
from jinja2 import Template
from loguru import logger
from luwu.core.models.complex.od.utils import label_map_util
from luwu.utils import cmd_util, file_util


class LuWuObjectDetector:
    def __init__(
        self,
        origin_dataset_path: str = "",
        tfrecord_dataset_path: str = "",
        label_map_path: str = "",
        model_save_path: str = "",
        batch_size: int = 8,
        steps: int = 2000,
        project_id: int = 0,
        **kwargs,
    ):
        """陆吾目标检测模型基类
        # TODO:暂时不增加验证集相关功能，后面再加

        Args:
            origin_dataset_path (str): 处理前的数据集路径
            tfrecord_dataset_path (str): 处理后的tfrecord数据集路径
            label_map_path (str): 目标检测类别映射表(pbtxt)
            model_save_path (str): 模型保存路径
            batch_size (int): mini batch 大小
            steps (int, optional): 训练steps数量. Defaults to 2000.
            project_id (int, optional): 项目编号. Defaults to 0.
        """
        self._call_code = ""
        self.project_id = project_id
        origin_dataset_path = file_util.abspath(origin_dataset_path)
        tfrecord_dataset_path = file_util.abspath(tfrecord_dataset_path)
        label_map_path = file_util.abspath(label_map_path)
        model_save_path = file_util.abspath(model_save_path)
        self.origin_dataset_path = origin_dataset_path
        # 当未给定处理后数据集的路径时，默认保存到原始数据集相同路径
        if tfrecord_dataset_path:
            # 区分给定的是文件夹还是文件。
            # 如果是文件夹，则需要生成tfrecord文件
            # 如果指定指定到文件，则直接使用指定文件，跳过生成步骤
            if os.path.isfile(tfrecord_dataset_path):
                self.tfrecord_dataset_file_path = tfrecord_dataset_path
                self.tfrecord_dataset_dir = os.path.dirname(tfrecord_dataset_path)
                self.need_generate_tfrecord = False
            else:
                self.tfrecord_dataset_dir = tfrecord_dataset_path
                self.tfrecord_dataset_file_path = os.path.join(
                    self.tfrecord_dataset_dir, "train.tfrecord"
                )
                self.need_generate_tfrecord = True
        else:
            self.tfrecord_dataset_dir = self.origin_dataset_path
            self.tfrecord_dataset_file_path = os.path.join(
                self.tfrecord_dataset_dir, "train.tfrecord"
            )
            self.need_generate_tfrecord = True
        # 当未给定pbtxt路径时，也默认保存到tfrecord相同目录下
        if label_map_path:
            if os.path.isfile(label_map_path):
                self.label_map_file_path = label_map_path
                self.label_map_dir = os.path.dirname(self.label_map_file_path)
                self.need_generate_label_map = False
            else:
                self.label_map_dir = label_map_path
                self.label_map_file_path = os.path.join(
                    self.label_map_dir, "label_map.pbtxt"
                )
                self.need_generate_label_map = True
        else:
            self.label_map_dir = self.tfrecord_dataset_dir
            self.label_map_file_path = os.path.join(
                self.label_map_dir, "label_map.pbtxt"
            )
            self.need_generate_label_map = True
        # 当未给定模型保存路径时，默认保存到处理后数据集相同路径
        if self.project_id:
            self.project_save_name = f"luwu-object-detection-project-{self.project_id}"
        else:
            self.project_save_name = f"luwu-object-detection-project"
        if model_save_path:
            self.project_save_path = os.path.join(
                model_save_path, self.project_save_name
            )
        else:
            self.project_save_path = os.path.join(
                self.tfrecord_dataset_dir, self.project_save_name
            )
        self.batch_size = batch_size
        self.steps = steps

    def preprocess_dataset(self):
        """对数据进行预处理"""
        raise NotImplementedError

    def train(self):
        """训练模型"""
        raise NotImplementedError

    def export_model(self):
        """导出训练好的模型"""
        raise NotImplementedError

    def run(self):
        """执行入口"""
        raise NotImplementedError


class LuWuTFModelsObjectDetector(LuWuObjectDetector):
    """目标检测器。
    基于 [https://github.com/tensorflow/models](https://github.com/tensorflow/models)
    """

    fine_tune_models_config_map = {
        "SSD ResNet50 V1 FPN 640x640 (RetinaNet50)": {
            "url": "http://download.tensorflow.org/models/object_detection/tf2/20200711/ssd_resnet50_v1_fpn_640x640_coco17_tpu-8.tar.gz",
            "speed": "46",
            "coco mAP": "34.3",
            "outputs": "Boxes",
            "template": "SSD_ResNet50_V1_FPN_640x640_(RetinaNet50).jinja",
        },
        "EfficientDet D0 512x512": {
            "url": "http://download.tensorflow.org/models/object_detection/tf2/20200711/efficientdet_d0_coco17_tpu-32.tar.gz",
            "speed": "39",
            "coco mAP": "33.6",
            "outputs": "Boxes",
            "template": "EfficientDet_D0_512x512.jinja",
        },
        "CenterNet Resnet101 V1 FPN 512x512": {
            "url": "http://download.tensorflow.org/models/object_detection/tf2/20200711/centernet_resnet101_v1_fpn_512x512_coco17_tpu-8.tar.gz",
            "speed": "34",
            "coco mAP": "34.2",
            "outputs": "Boxes",
            "template": "CenterNet_Resnet101_V1_FPN_512x512.jinja",
        },
        "CenterNet HourGlass104 512x512": {
            "url": "http://download.tensorflow.org/models/object_detection/tf2/20200713/centernet_hg104_512x512_coco17_tpu-8.tar.gz",
            "speed": "70",
            "coco mAP": "41.9",
            "outputs": "Boxes",
            "template": "CenterNet_HourGlass104_512x512.jinja",
        },
    }

    def __init__(
        self,
        origin_dataset_path: str = "",
        tfrecord_dataset_path: str = "",
        label_map_path: str = "",
        do_fine_tune: bool = True,
        fine_tune_checkpoint_path: str = "",
        fine_tune_model_name: str = "",
        model_save_path: str = "",
        batch_size: int = 8,
        steps: int = 2000,
        project_id: int = 0,
        **kwargs,
    ):
        """陆吾目标检测模型基类
        # TODO:暂时不增加验证集相关功能，后面再加

        Args:
            origin_dataset_path (str): 处理前的数据集路径
            tfrecord_dataset_path (str): 处理后的tfrecord数据集路径
            label_map_path (str): 目标检测类别映射表(pbtxt)
            do_fine_tune (bool): 是否在预训练模型的基础上进行微调
            fine_tune_checkpoint_path (str): 预训练权重路径
            fine_tune_model_name (str): 预训练模型名称
            model_save_path (str): 模型保存路径
            batch_size (int): mini batch 大小
            steps (int, optional): 训练steps数量. Defaults to 2000.
            project_id (int, optional): 项目编号. Defaults to 0.
        """
        super().__init__(
            origin_dataset_path=origin_dataset_path,
            tfrecord_dataset_path=tfrecord_dataset_path,
            label_map_path=label_map_path,
            model_save_path=model_save_path,
            batch_size=batch_size,
            steps=steps,
            project_id=project_id,
            **kwargs,
        )
        self.do_fine_tune = do_fine_tune
        fine_tune_checkpoint_path = file_util.abspath(fine_tune_checkpoint_path)
        self.fine_tune_checkpoint_path = fine_tune_checkpoint_path
        self.fine_tune_model_name = fine_tune_model_name
        if self.fine_tune_model_name not in self.fine_tune_models_config_map:
            raise Exception(
                f"暂不支持的 object detection model! {self.fine_tune_model_name}"
            )

    def preprocess_dataset(self):
        """对数据集进行预处理，并定义pipeline.config"""
        # 生成 label_map.pbtxt
        if self.need_generate_label_map:
            logger.info("遍历数据集，生成 label_map.pbtxt ...")
            if not os.path.exists(self.origin_dataset_path):
                raise FileNotFoundError("origin_dataset_path 未指定！")
            label_map_util.create_label_map(
                self.origin_dataset_path, self.label_map_file_path
            )
            logger.info(f"label_map 文件已保存到 {self.label_map_file_path}.")
        else:
            logger.info(f"label_map 文件已存在，路径为{self.label_map_file_path}，跳过！")

        # 生成 tfrecord
        if self.need_generate_tfrecord:
            logger.info("遍历数据集，生成 tfrecord ...")
            python_execute_path = cmd_util.get_python_execute_path()
            script_path = os.path.join(
                os.path.dirname(__file__), "utils", "generate_tfrecord.py"
            )
            csv_path = os.path.join(self.tfrecord_dataset_dir, "tmp.csv")
            cmd = f"{python_execute_path} {script_path} -x {self.origin_dataset_path} -l {self.label_map_file_path} -o {self.tfrecord_dataset_file_path} -c {csv_path}"
            cmd_util.run_cmd(cmd)
            if os.path.exists(csv_path):
                os.remove(csv_path)
        else:
            logger.info(f"tfrecord 文件已存在，路径为{self.tfrecord_dataset_file_path}，跳过！")

        if self.do_fine_tune:
            # 下载预训练权重
            url = self.fine_tune_models_config_map[self.fine_tune_model_name]["url"]
            download_dir_name = url.split("/")[-1].split(".")[0]
            cache_dir_path = os.path.expanduser(
                "~/.luwu/tensorflow-models/object-detection/"
            )
            if self.fine_tune_checkpoint_path:
                # 如果给定路径指向了checkpoint文件夹下ckpt-0.index文件
                if self.fine_tune_checkpoint_path.endswith(".index"):
                    file_path = self.fine_tune_checkpoint_path
                    self.fine_tune_checkpoint_path = (
                        self.fine_tune_checkpoint_path.rstrip(".index")
                    )
                # 指向checkpoint文件夹
                elif (
                    (
                        self.fine_tune_checkpoint_path.endswith("checkpoint/")
                        or self.fine_tune_checkpoint_path.endswith("checkpoint")
                    )
                    and os.path.exists(self.fine_tune_checkpoint_path)
                    and os.path.isdir(self.fine_tune_checkpoint_path)
                ):
                    file_path = os.path.join(
                        self.fine_tune_checkpoint_path, "ckpt-0.index"
                    )
                    self.fine_tune_checkpoint_path = os.path.join(
                        self.fine_tune_checkpoint_path, "ckpt-0"
                    )
                # 默认的checkpoint路径
                else:
                    self.fine_tune_checkpoint_path = os.path.join(
                        cache_dir_path,
                        download_dir_name,
                        "checkpoint/ckpt-0",
                    )
                    file_path = self.fine_tune_checkpoint_path + ".index"
            else:
                self.fine_tune_checkpoint_path = os.path.join(
                    cache_dir_path,
                    download_dir_name,
                    "checkpoint/ckpt-0",
                )
                file_path = self.fine_tune_checkpoint_path + ".index"
            # 检查文件是否存在
            if os.path.exists(file_path):
                logger.info(f"预训练权重已存在 {self.fine_tune_checkpoint_path}，跳过！")
            else:
                tf.keras.utils.get_file(
                    download_dir_name, url, untar=True, cache_subdir=cache_dir_path
                )
                logger.info("预训练权重下载完成！")
        else:
            self.fine_tune_checkpoint_path = ""
            logger.info("不使用预训练权重！")

        # 创建项目文件夹结构
        # 根文件夹，如果存在则会抛出异常
        if os.path.exists(self.project_save_path):
            while True:
                text = input(
                    f"目录 {self.project_save_path} 已存在，请更换目录或者确认清空该文件夹！确认清空？[Y/N]"
                )
                text = text.lower().strip()
                if text == "y":
                    cmd = f"rm -rf {self.project_save_path} -y"
                    cmd_util.run_cmd(cmd)
                    os.makedirs(self.project_save_path)
                    break
                elif text == "n":
                    logger.info("请重选选择模型保存的目录，程序已退出。")
                    exit(-1)
                else:
                    continue
        else:
            os.makedirs(self.project_save_path)
        # 训练checkpoint保存路径
        self.train_checkpoint_path = os.path.join(
            self.project_save_path, "train_models"
        )
        os.makedirs(self.train_checkpoint_path, exist_ok=True)
        # 导出模型（SavedModel）保存路径
        self.export_model_path = os.path.join(self.project_save_path, "exported-models")
        os.makedirs(self.export_model_path, exist_ok=True)

        # 创建pipeline.config
        self.generate_pipeline_config()
        logger.info("pipeline.config 生成完毕！")

    def render_template(self, template_path, params) -> str:
        """读取并使用给定的参数渲染模板，返回渲染后的内容

        Args:
            template_path (str): 模板路径
            params (dict)): 填充模板的参数集

        Return:
            str: 渲染之后的template内容
        """
        with open(template_path, "r") as f:
            text = f.read()
        template = Template(text)
        content = template.render(**params)
        return content

    def generate_pipeline_config(self, *args, **kwargs):
        params = {}
        label_map_dict = label_map_util.get_label_map_dict(self.label_map_file_path)
        num_classes = len(label_map_dict)
        # TODO: 优化初始步数生成规则，不再使用默认值
        if self.fine_tune_model_name in (
            "SSD ResNet50 V1 FPN 640x640 (RetinaNet50)",
            "EfficientDet D0 512x512",
            "CenterNet Resnet101 V1 FPN 512x512",
            "CenterNet HourGlass104 512x512",
        ):
            params["num_classes"] = num_classes
            params["batch_size"] = self.batch_size
            params["num_steps"] = self.steps
            params["fine_tune_checkpoint"] = self.fine_tune_checkpoint_path
            params["label_map_path"] = self.label_map_file_path
            params["train_input_path"] = self.tfrecord_dataset_file_path
            params["eval_input_path"] = self.tfrecord_dataset_file_path
        else:
            raise Exception(
                f"暂不支持的 object detection model! {self.fine_tune_model_name}"
            )
        template_name = self.fine_tune_models_config_map[self.fine_tune_model_name][
            "template"
        ]
        template_path = template_path = os.path.join(
            os.path.dirname(__file__), f"templates/pipeline/{template_name}"
        )
        content = self.render_template(template_path, params)
        self.train_pipeline_config_path = os.path.join(
            self.train_checkpoint_path, "pipeline.config"
        )
        with open(self.train_pipeline_config_path, "w") as f:
            f.write(content)

    def train(self):
        logger.info("开始训练...")
        python_execute_path = cmd_util.get_python_execute_path()
        cmd = f"{python_execute_path} -m object_detection.model_main_tf2 --model_dir={self.train_checkpoint_path} --pipeline_config_path={self.train_pipeline_config_path}"
        cmd_util.run_cmd(cmd)
        logger.info("训练完成！")

    def export_model(self):
        """将训练好的模型导出为pb格式"""
        logger.info("正在导出模型...")
        python_execute_path = cmd_util.get_python_execute_path()
        cmd = f"{python_execute_path} -m object_detection.exporter_main_v2 --input_type image_tensor --pipeline_config_path {self.train_pipeline_config_path} --trained_checkpoint_dir {self.train_checkpoint_path} --output_directory {self.export_model_path}"
        cmd_util.run_cmd(cmd)

        # 复制一份label_map.pbtxt到该目录
        logger.info("导出 label_map.pbtxt...")
        target_path = os.path.join(self.export_model_path, "label_map.pbtxt")
        cmd = f"cp {self.label_map_file_path} {target_path}"
        cmd_util.run_cmd(cmd)

        logger.info("导出测试代码...")
        if os.path.exists(self.origin_dataset_path):
            filenames = glob(os.path.join(self.origin_dataset_path, "*.jpg"))
            if len(filenames):
                eval_image_path = filenames[0]
            else:
                eval_image_path = ""
        else:
            eval_image_path = ""
        params = {
            "path_to_saved_model": os.path.join(self.export_model_path, "saved_model"),
            "path_to_label_map": target_path,
            "path_to_eval_image": eval_image_path,
        }
        template_path = template_path = os.path.join(
            os.path.dirname(__file__), "templates/project/eval.jinja"
        )
        content = self.render_template(template_path, params)
        target_path = os.path.join(self.project_save_path, "eval.py")
        with open(target_path, "w") as f:
            f.write(content)
        logger.info("导出完成！")

    def run(self):
        # 预处理
        self.preprocess_dataset()
        # 训练
        self.train()
        # 导出模型
        self.export_model()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-21
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/luwu/core/models/classifier/preset/pre_trained.py
# @Desc         : 封装tf.keras里设置的预训练模型，并对外提供支持
import os

import tensorflow as tf
from jinja2 import Template
from luwu.core.models.classifier import LuwuImageClassifier


class LuwuPreTrainedImageClassifier(LuwuImageClassifier):
    def __init__(self, net_name, *args, do_fine_tune=False, **kwargs):
        super(LuwuPreTrainedImageClassifier, self).__init__(*args, **kwargs)
        self.net_name = net_name
        self.do_fine_tune = do_fine_tune

    def build_model(self):
        pre_trained_net: tf.keras.Model = getattr(
            tf.keras.applications, self.net_name
        )()
        pre_trained_net.trainable = False
        # 记录densenet
        self.pre_trained_net = pre_trained_net
        # model = tf.keras.Sequential(
        #     [
        #         pre_trained_net,
        #         tf.keras.layers.Flatten(),
        #         tf.keras.layers.Dense(120, activation="relu"),
        #         tf.keras.layers.Dropout(0.3),
        #         tf.keras.layers.Dense(len(self.classes_num_dict), activation="softmax"),
        #     ]
        # )
        model = tf.keras.Sequential(
            [
                pre_trained_net,
                tf.keras.layers.Flatten(),
                tf.keras.layers.Dense(len(self.classes_num_dict), activation="softmax"),
            ]
        )
        model.compile(
            optimizer=tf.keras.optimizers.Adam(),
            loss=tf.keras.losses.categorical_crossentropy,
            metrics=["accuracy"],
        )
        self.model = model
        return model

    def train(self):
        # callbacks
        checkpoint = tf.keras.callbacks.ModelCheckpoint(
            self.model_save_path, monitor="val_accuracy", save_best_only=True
        )
        if self.do_fine_tune:
            pre_train_epochs = int(self.epochs / 2)
        else:
            pre_train_epochs = 0
        train_epochs = self.epochs - pre_train_epochs
        # 训练
        if pre_train_epochs:
            self.model.fit(
                self.train_dataset.for_fit(),
                initial_epoch=0,
                epochs=pre_train_epochs,
                steps_per_epoch=self.train_dataset.steps,
                validation_data=self.dev_dataset.for_fit(),
                validation_steps=self.dev_dataset.steps,
                callbacks=[
                    checkpoint,
                ],
            )
        if train_epochs:
            self.pre_trained_net.trainable = True
            self.model.fit(
                self.train_dataset.for_fit(),
                initial_epoch=pre_train_epochs,
                epochs=self.epochs,
                steps_per_epoch=self.train_dataset.steps,
                validation_data=self.dev_dataset.for_fit(),
                validation_steps=self.dev_dataset.steps,
                callbacks=[
                    checkpoint,
                ],
            )

    def get_call_code(self):
        """返回模型定义和模型调用的代码"""
        if not self._call_code:
            template_path = os.path.join(
                os.path.dirname(__file__), "templates/LuwuPreTrainedImageClassifier.txt"
            )
            with open(template_path, "r") as f:
                text = f.read()
            data = {
                "net_name": self.net_name,
                "num_classes": len(self.classes_num_dict),
                "num_classes_map": str(self.classes_num_dict_rev),
                "image_size": self.image_size,
                "model_path": self.model_save_path,
                "data_preprocess_template": self.train_dataset.generate_preprocess_code(),
            }
            template = Template(text)
            code = template.render(**data)
            self._call_code = code
        return self._call_code

    def save_code(self):
        """导出模型定义和模型调用的代码"""
        code = self.get_call_code()
        code_file_name = "luwu-code.py"
        code_path = os.path.join(self.project_save_path, code_file_name)
        with open(code_path, "w") as f:
            f.write(code)


class LuwuLeNetImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, *args, image_size=32, **kwargs):
        kwargs["net_name"] = "LeNet"
        super(LuwuLeNetImageClassifier, self).__init__(*args, **kwargs)
        self.image_size = image_size

    def build_model(self):
        model = tf.keras.Sequential(
            [
                tf.keras.layers.Conv2D(6, (5, 5), padding="same"),
                # 添加BN层，将数据调整为均值0，方差1
                tf.keras.layers.BatchNormalization(),
                # 最大池化层，池化后图片长宽减半
                tf.keras.layers.MaxPooling2D((2, 2), 2),
                # relu激活层
                tf.keras.layers.ReLU(),
                # 第二个卷积层
                tf.keras.layers.Conv2D(16, (5, 5)),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.MaxPooling2D((2, 2), 2),
                tf.keras.layers.ReLU(),
                # 将节点展平为(None,-1)的形式，以作为全连接层的输入
                tf.keras.layers.Flatten(),
                # 第一个全连接层，120个节点
                tf.keras.layers.Dense(120),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.ReLU(),
                # 第二个全连接层
                tf.keras.layers.Dense(84),
                tf.keras.layers.BatchNormalization(),
                tf.keras.layers.ReLU(),
                # 输出层，使用softmax激活
                tf.keras.layers.Dense(len(self.classes_num_dict), activation="softmax"),
            ]
        )
        model.compile(
            optimizer=tf.keras.optimizers.Adam(),
            loss=tf.keras.losses.categorical_crossentropy,
            metrics=["accuracy"],
        )
        self.model = model
        return model

    def train(self):
        # callbacks
        checkpoint = tf.keras.callbacks.ModelCheckpoint(
            self.model_save_path, monitor="val_accuracy", save_best_only=True
        )
        self.model.fit(
            self.train_dataset.for_fit(),
            epochs=self.epochs,
            steps_per_epoch=self.train_dataset.steps,
            validation_data=self.dev_dataset.for_fit(),
            validation_steps=self.dev_dataset.steps,
            callbacks=[
                checkpoint,
            ],
        )


class LuwuDenseNet121ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="DenseNet121", **kwargs):
        super(LuwuDenseNet121ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuDenseNet169ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="DenseNet169", **kwargs):
        super(LuwuDenseNet169ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuDenseNet201ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="DenseNet201", **kwargs):
        super(LuwuDenseNet201ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuVGG16ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="VGG16", **kwargs):
        super(LuwuVGG16ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuVGG19ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="VGG19", **kwargs):
        super(LuwuVGG19ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuMobileNetImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="MobileNet", **kwargs):
        super(LuwuMobileNetImageClassifier, self).__init__(net_name, **kwargs)


class LuwuMobileNetV2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="MobileNetV2", **kwargs):
        super(LuwuMobileNetV2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuInceptionResNetV2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="InceptionResNetV2", **kwargs):
        super(LuwuInceptionResNetV2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuInceptionV3ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="InceptionV3", **kwargs):
        super(LuwuInceptionV3ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuNASNetMobileImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="NASNetMobile", **kwargs):
        super(LuwuNASNetMobileImageClassifier, self).__init__(net_name, **kwargs)


class LuwuNASNetLargeImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="NASNetLarge", **kwargs):
        super(LuwuNASNetLargeImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet50ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet50", **kwargs):
        super(LuwuResNet50ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet50V2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet50V2", **kwargs):
        super(LuwuResNet50V2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet101ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet101", **kwargs):
        super(LuwuResNet101ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet101V2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet101V2", **kwargs):
        super(LuwuResNet101V2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet152ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet152", **kwargs):
        super(LuwuResNet152ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuResNet152V2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="ResNet152V2", **kwargs):
        super(LuwuResNet152V2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuMobileNetV3SmallImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="MobileNetV3Small", **kwargs):
        super(LuwuMobileNetV3SmallImageClassifier, self).__init__(net_name, **kwargs)


class LuwuMobileNetV3LargeImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="MobileNetV3Large", **kwargs):
        super(LuwuMobileNetV3LargeImageClassifier, self).__init__(net_name, **kwargs)


class LuwuXceptionImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="Xception", **kwargs):
        super(LuwuXceptionImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB0ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB0", **kwargs):
        super(LuwuEfficientNetB0ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB1ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB1", **kwargs):
        super(LuwuEfficientNetB1ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB2ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB2", **kwargs):
        super(LuwuEfficientNetB2ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB3ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB3", **kwargs):
        super(LuwuEfficientNetB3ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB4ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB4", **kwargs):
        super(LuwuEfficientNetB4ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB5ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB5", **kwargs):
        super(LuwuEfficientNetB5ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB6ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB6", **kwargs):
        super(LuwuEfficientNetB6ImageClassifier, self).__init__(net_name, **kwargs)


class LuwuEfficientNetB7ImageClassifier(LuwuPreTrainedImageClassifier):
    def __init__(self, net_name="EfficientNetB7", **kwargs):
        super(LuwuEfficientNetB7ImageClassifier, self).__init__(net_name, **kwargs)



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-30
# @FilePath     : /LuWu/luwu/backend/model.py
# @Desc         :
import json
from datetime import datetime

from flask_sqlalchemy import SQLAlchemy

db = SQLAlchemy()


class TrainProject(db.Model):
    id = db.Column(db.Integer, primary_key=True, autoincrement=True)
    params_text = db.Column(db.Text, nullable=False, default="", comment="参数json")
    code = db.Column(db.Text, nullable=False, default="", comment="生成的调用代码")
    model_name = db.Column(db.String(100), nullable=False, default="", comment="模型名称")
    status = db.Column(
        db.Integer,
        nullable=False,
        default=0,
        comment="处理状态 0-未处理，1-等待处理，2-处理中，3-处理完成，4-处理失败，5-手动终止，6-异常退出",
    )
    deleted = db.Column(db.Integer, nullable=False, default=0, comment="是否删除，1-删除")
    addtime = db.Column(db.Integer, nullable=False, default=0, comment="创建时间")

    @property
    def status_text(self):
        status_dict = {
            0: "未处理",
            1: "等待处理",
            2: "处理中",
            3: "处理完成",
            4: "处理失败",
            5: "手动终止",
            6: "异常退出",
        }
        return status_dict[self.status]

    @property
    def params(self):
        return json.loads(self.params_text)

    @params.setter
    def params(self, value):
        self.params_text = json.dumps(value)

    def to_dict(self):
        data = {
            "id": self.id,
            "params": self.params,
            "params_text": self.params_text,
            "code": self.code,
            "model_name": self.model_name,
            "deleted": self.deleted,
            "status": self.status,
            "status_text": self.status_text,
            "addtime": datetime.fromtimestamp(self.addtime).strftime(
                "%Y-%m-%d %H:%M:%S"
            ),
        }
        return data

    def add(self):
        db.session.add(self)
        db.session.commit()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-21
# @Author       : AaronJny
# @LastEditTime : 2021-03-12
# @FilePath     : /LuWu/luwu/core/models/classifier/preset/__init__.py
# @Desc         :
from luwu.core.models.classifier.preset.pre_trained import (
    LuwuLeNetImageClassifier,
    LuwuDenseNet121ImageClassifier,
    LuwuDenseNet169ImageClassifier,
    LuwuDenseNet201ImageClassifier,
    LuwuVGG16ImageClassifier,
    LuwuVGG19ImageClassifier,
    LuwuMobileNetImageClassifier,
    LuwuMobileNetV2ImageClassifier,
    LuwuInceptionResNetV2ImageClassifier,
    LuwuInceptionV3ImageClassifier,
    LuwuNASNetMobileImageClassifier,
    LuwuNASNetLargeImageClassifier,
    LuwuResNet50ImageClassifier,
    LuwuResNet50V2ImageClassifier,
    LuwuResNet101ImageClassifier,
    LuwuResNet101V2ImageClassifier,
    LuwuResNet152ImageClassifier,
    LuwuResNet152V2ImageClassifier,
    LuwuMobileNetV3SmallImageClassifier,
    LuwuMobileNetV3LargeImageClassifier,
    LuwuXceptionImageClassifier,
    LuwuEfficientNetB0ImageClassifier,
    LuwuEfficientNetB1ImageClassifier,
    LuwuEfficientNetB2ImageClassifier,
    LuwuEfficientNetB3ImageClassifier,
    LuwuEfficientNetB4ImageClassifier,
    LuwuEfficientNetB5ImageClassifier,
    LuwuEfficientNetB6ImageClassifier,
    LuwuEfficientNetB7ImageClassifier,
)

__all__ = [
    "LuwuPreTrainedImageClassifier",
    "LuwuLeNetImageClassifier",
    "LuwuDenseNet121ImageClassifier",
    "LuwuDenseNet169ImageClassifier",
    "LuwuDenseNet201ImageClassifier",
    "LuwuVGG16ImageClassifier",
    "LuwuVGG19ImageClassifier",
    "LuwuMobileNetImageClassifier",
    "LuwuMobileNetV2ImageClassifier",
    "LuwuInceptionResNetV2ImageClassifier",
    "LuwuInceptionV3ImageClassifier",
    "LuwuNASNetMobileImageClassifier",
    "LuwuNASNetLargeImageClassifier",
    "LuwuResNet50ImageClassifier",
    "LuwuResNet50V2ImageClassifier",
    "LuwuResNet101ImageClassifier",
    "LuwuResNet101V2ImageClassifier",
    "LuwuResNet152ImageClassifier",
    "LuwuResNet152V2ImageClassifier",
    "LuwuMobileNetV3SmallImageClassifier",
    "LuwuMobileNetV3LargeImageClassifier",
    "LuwuXceptionImageClassifier",
    "LuwuEfficientNetB0ImageClassifier",
    "LuwuEfficientNetB1ImageClassifier",
    "LuwuEfficientNetB2ImageClassifier",
    "LuwuEfficientNetB3ImageClassifier",
    "LuwuEfficientNetB4ImageClassifier",
    "LuwuEfficientNetB5ImageClassifier",
    "LuwuEfficientNetB6ImageClassifier",
    "LuwuEfficientNetB7ImageClassifier",
]



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-29
# @FilePath     : /LuWu/luwu/run.py
# @Desc         :
import time
import traceback
from multiprocessing import Process
import os
import sys

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

import tensorflow as tf
from loguru import logger

# 引入顺序不能变动，必须先执行此段代码，才能引入luwu下的模块
gpus = tf.config.experimental.list_physical_devices(device_type="GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)

from luwu.backend import app
from luwu.scripts import scheduler
from luwu.backend.config import Config


def init_luwu_dir():
    dir_name = Config.LUWU_DIR
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)


def run():
    while True:
        processes = []
        init_luwu_dir()
        try:
            # 启动Web服务
            logger.info("正在启动 Web 服务进程...")
            web_process = Process(target=app.run)
            processes.append(web_process)
            web_process.start()

            time.sleep(20)

            # 启动调度进程
            logger.info("正在启动任务调度进程...")
            scheduler_process = Process(target=scheduler.run)
            processes.append(scheduler_process)
            scheduler_process.start()

            for process in processes:
                process.join()

        except KeyboardInterrupt:
            logger.info("收到终止信号，正在关闭所有进程...")
            for process in processes:
                if process.is_alive():
                    process.terminate()
                if process.is_alive():
                    process.kill()
            logger.info("关闭完成！结束程序！")
            break
        except Exception as e:
            logger.error(e)
            logger.error(traceback.format_exc())
            # 收到终止信号或抛出异常时，关闭所有进程后再退出
            logger.info("出现异常！正在关闭所有进程...")
            for process in processes:
                if process.is_alive():
                    process.terminate()
                if process.is_alive():
                    process.kill()
            logger.info("关闭完成！正在重试...！")
            time.sleep(10)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-12
# @FilePath     : /LuWu/luwu/utils/file_util.py
# @Desc         :
import os


def abspath(filepath):
    if filepath:
        return os.path.abspath(os.path.expanduser(filepath))
    else:
        return ""


def mkdirs(dirpath):
    os.makedirs(dirpath, exist_ok=True)



# -*- coding: utf-8 -*-
# @Date         : 2020-12-30
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/luwu/core/models/classifier/__init__.py
# @Desc         :
import os
import random

import tensorflow as tf
from loguru import logger
from luwu.core.preprocess.data.data_generator import ImageClassifierDataGnenrator
from luwu.core.preprocess.image.load import (
    read_classify_dataset_from_dir,
    write_tfrecords_to_target_path,
)
from luwu.utils import file_util
import numpy as np


class LuwuImageClassifier:
    def __init__(
        self,
        origin_dataset_path: str = "",
        target_dataset_path: str = "",
        model_save_path: str = "",
        validation_split: float = 0.2,
        batch_size: int = 32,
        epochs: int = 30,
        project_id: int = 0,
        image_size: int = 224,
        **kwargs,
    ):
        """
        Args:
            origin_dataset_path (str): 处理前的数据集路径
            target_dataset_path (str): 处理后的数据集路径
            model_save_path (str): 模型保存路径
            validation_split (float): 验证集切割比例
            batch_size (int): mini batch 大小
            epochs (int): 训练epoch数
            project_id (int): 训练项目编号
        """
        self._call_code = ""
        self.project_id = project_id
        origin_dataset_path = file_util.abspath(origin_dataset_path)
        target_dataset_path = file_util.abspath(target_dataset_path)
        model_save_path = file_util.abspath(model_save_path)
        self.image_size = image_size
        self.origin_dataset_path = origin_dataset_path
        # 当未给定处理后数据集的路径时，默认保存到原始数据集相同路径
        if target_dataset_path:
            self.target_dataset_path = target_dataset_path
        else:
            self.target_dataset_path = origin_dataset_path
        # 当未给定模型保存路径时，默认保存到处理后数据集相同路径
        if self.project_id:
            self.project_save_name = f"luwu-classification-project-{self.project_id}"
        else:
            self.project_save_name = f"luwu-classification-project"
        if model_save_path:
            self.project_save_path = os.path.join(
                model_save_path, self.project_save_name
            )
        else:
            self.project_save_path = os.path.join(
                self.target_dataset_path, self.project_save_name
            )
        self.model_save_path = os.path.join(self.project_save_path, "best_weights.h5")
        self.validation_split = validation_split
        self.batch_size = batch_size
        self.epochs = epochs
        file_util.mkdirs(self.project_save_path)
        file_util.mkdirs(self.target_dataset_path)

    def build_model(self) -> tf.keras.Model:
        """构建模型

        Raises:
            NotImplementedError: 待实现具体方法
        """
        raise NotImplementedError

    def preprocess_dataset(self):
        """对数据集进行预处理"""
        # 读取原始数据
        data, classes_num_dict = read_classify_dataset_from_dir(
            self.origin_dataset_path
        )
        # 类别->编号的映射
        self.classes_num_dict = classes_num_dict
        # 编号->类别的映射
        self.classes_num_dict_rev = {
            value: key for key, value in self.classes_num_dict.items()
        }
        # 先判断tfrecord是否存在
        self.target_train_dataset_path = os.path.join(
            self.target_dataset_path, "train_dataset"
        )
        self.target_dev_dataset_path = os.path.join(
            self.target_dataset_path, "dev_dataset"
        )
        # if (
        #     os.path.exists(self.target_train_dataset_path)
        #     and os.path.exists(self.target_dev_dataset_path)
        #     and os.path.isfile(self.target_train_dataset_path)
        #     and os.path.isfile(self.target_dev_dataset_path)
        # ):
        #     logger.info("TFRecord数据集已存在。跳过！")
        # else:
        # 切分数据
        total = len(data)
        dev_nums = int(total * self.validation_split)
        dev_data = random.sample(data, dev_nums)
        train_data = list(set(data) - set(dev_data))
        np.random.shuffle(train_data)
        np.random.shuffle(dev_data)
        del data

        # 制作tfrecord数据集
        write_tfrecords_to_target_path(
            train_data, len(classes_num_dict), self.target_train_dataset_path
        )
        write_tfrecords_to_target_path(
            dev_data, len(classes_num_dict), self.target_dev_dataset_path
        )
        # 读取tfrecord数据集
        self.train_dataset = ImageClassifierDataGnenrator(
            self.target_train_dataset_path, batch_size=self.batch_size
        )
        self.dev_dataset = ImageClassifierDataGnenrator(
            self.target_dev_dataset_path, batch_size=self.batch_size, shuffle=False
        )

    def train(self):
        # callbacks
        checkpoint = tf.keras.callbacks.ModelCheckpoint(
            self.model_save_path, monitor="val_accuracy", save_best_only=True
        )
        # 训练
        self.model.fit(
            self.train_dataset.for_fit(),
            epochs=self.epochs,
            steps_per_epoch=self.train_dataset.steps,
            validation_data=self.dev_dataset.for_fit(),
            validation_steps=self.dev_dataset.steps,
            callbacks=[
                checkpoint,
            ],
        )

    def run(self):
        # 预处理数据集
        logger.info("正在预处理数据集...")
        self.preprocess_dataset()
        # 构建模型
        logger.info("正在构建模型...")
        self.build_model()
        # 训练模型
        logger.info("开始训练...")
        self.train()
        # 导出代码
        logger.info("导出代码...")
        self.save_code()
        logger.info("Done.")

    def generator_train_code(self):
        """导出模型定义和训练代码"""
        raise NotImplementedError

    def get_call_code(self):
        """返回模型定义和模型调用的代码"""
        if self._call_code:
            return self._call_code
        else:
            raise NotImplementedError

    def save_code(self):
        """导出模型定义和模型调用的代码"""
        raise NotImplementedError


from luwu.core.models.classifier.preset import (
    LuwuLeNetImageClassifier,
    LuwuDenseNet121ImageClassifier,
    LuwuDenseNet169ImageClassifier,
    LuwuDenseNet201ImageClassifier,
    LuwuInceptionResNetV2ImageClassifier,
    LuwuInceptionV3ImageClassifier,
    LuwuMobileNetImageClassifier,
    LuwuMobileNetV2ImageClassifier,
    LuwuMobileNetV3LargeImageClassifier,
    LuwuMobileNetV3SmallImageClassifier,
    LuwuNASNetLargeImageClassifier,
    LuwuNASNetMobileImageClassifier,
    LuwuResNet50ImageClassifier,
    LuwuResNet50V2ImageClassifier,
    LuwuResNet101ImageClassifier,
    LuwuResNet101V2ImageClassifier,
    LuwuResNet152ImageClassifier,
    LuwuResNet152V2ImageClassifier,
    LuwuVGG16ImageClassifier,
    LuwuVGG19ImageClassifier,
    LuwuXceptionImageClassifier,
    LuwuEfficientNetB0ImageClassifier,
    LuwuEfficientNetB1ImageClassifier,
    LuwuEfficientNetB2ImageClassifier,
    LuwuEfficientNetB3ImageClassifier,
    LuwuEfficientNetB4ImageClassifier,
    LuwuEfficientNetB5ImageClassifier,
    LuwuEfficientNetB6ImageClassifier,
    LuwuEfficientNetB7ImageClassifier,
)

__all__ = [
    "LuwuImageClassifier",
    "LuwuPreTrainedImageClassifier",
    "LuwuLeNetImageClassifier",
    "LuwuDenseNet121ImageClassifier",
    "LuwuDenseNet169ImageClassifier",
    "LuwuDenseNet201ImageClassifier",
    "LuwuVGG16ImageClassifier",
    "LuwuVGG19ImageClassifier",
    "LuwuMobileNetImageClassifier",
    "LuwuMobileNetV2ImageClassifier",
    "LuwuInceptionResNetV2ImageClassifier",
    "LuwuInceptionV3ImageClassifier",
    "LuwuNASNetMobileImageClassifier",
    "LuwuNASNetLargeImageClassifier",
    "LuwuResNet50ImageClassifier",
    "LuwuResNet50V2ImageClassifier",
    "LuwuResNet101ImageClassifier",
    "LuwuResNet101V2ImageClassifier",
    "LuwuResNet152ImageClassifier",
    "LuwuResNet152V2ImageClassifier",
    "LuwuMobileNetV3SmallImageClassifier",
    "LuwuMobileNetV3LargeImageClassifier",
    "LuwuXceptionImageClassifier",
    "LuwuEfficientNetB0ImageClassifier",
    "LuwuEfficientNetB1ImageClassifier",
    "LuwuEfficientNetB2ImageClassifier",
    "LuwuEfficientNetB3ImageClassifier",
    "LuwuEfficientNetB4ImageClassifier",
    "LuwuEfficientNetB5ImageClassifier",
    "LuwuEfficientNetB6ImageClassifier",
    "LuwuEfficientNetB7ImageClassifier",
]



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-29
# @FilePath     : /LuWu/luwu/backend/views.py
# @Desc         :
from flask import Blueprint
from flask import render_template

main_blueprint = Blueprint("main_blueprint", __name__)


@main_blueprint.route("/", defaults={"path": ""})
@main_blueprint.route("/home/<path:path>")
def catch_all(path):
    return render_template("index.html")


# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /LuWu/luwu/scripts/train_project.py
# @Desc         :
import sys
import os

sys.path.append(
    os.path.abspath(os.path.join(os.path.join(os.path.dirname(__file__), ".."), ".."))
)
import tensorflow as tf

# 引入顺序不能变动，必须先执行此段代码，才能引入luwu下的模块
gpus = tf.config.experimental.list_physical_devices(device_type="GPU")
for gpu in gpus:
    tf.config.experimental.set_memory_growth(gpu, True)


import fire
from loguru import logger
from luwu.scripts.utils import (
    get_project_by_id,
    load_model_class,
    update_project_code,
    update_project_status,
)


def train(project):
    update_project_status(project["id"], 2)
    model_name = project["model_name"]
    model_class = load_model_class(model_name)
    model = model_class(project_id=project["id"], **project["params"])
    model.run()
    code = model.get_call_code()
    update_project_code(project["id"], code)


@logger.catch
def run(project_id):
    project = get_project_by_id(project_id)
    try:
        train(project)
        update_project_status(project["id"], 2)
        logger.info("处理完成！")
    except Exception as e:
        update_project_status(project["id"], 6)
        logger.error("处理异常！")
        raise e


if __name__ == "__main__":
    fire.Fire(run)



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-30
# @FilePath     : /LuWu/luwu/backend/__init__.py
# @Desc         :
import functools
import traceback

from flask import jsonify
from loguru import logger


def status_code_wrapper(fix=True):
    """
    一个装饰器，用于自动为接口加上状态码和错误捕捉

    Args:
        fix: 是否需要对返回结果进行封装
    """

    def decorate(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            ret = {"code": 0, "msg": "请求成功", "data": []}
            try:
                data = func(*args, **kwargs)
                if not fix:
                    ret = data
                elif data:
                    ret["data"] = data
            except Exception as e:
                logger.error(e)
                traceback.print_exc()
                ret["code"] = -1
                ret["msg"] = str(e)
            return jsonify(ret)

        return wrap

    return decorate



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-29
# @FilePath     : /LuWu/luwu/scripts/utils.py
# @Desc         :
import importlib
import re
from urllib.parse import urljoin

import requests
from loguru import logger
from luwu.backend.config import Config

HOST = f"http://localhost:{Config.PORT}/"


def read_train_projects():
    """获取项目列表"""
    url = urljoin(HOST, "/api/v1/project/list/")
    resp = requests.get(url, timeout=30)
    if resp.status_code != 200:
        raise Exception("接口请求失败！")
    data = resp.json()
    if data["code"] != 0:
        raise Exception("接口请求失败！")
    return data["data"]


def update_project_status(xid, status):
    """更新项目状态

    Args:
        xid (int): 项目编号
        status (int): 状态编码
    """
    url = urljoin(HOST, f"/api/v1/project/{xid}/status/update/")
    data = {"status": status}
    logger.debug(f"更新状态 {xid} {status}")
    resp = requests.post(url, json=data)
    logger.debug(resp.json())


def update_project_code(xid, code):
    """更新项目状态

    Args:
        xid (int): 项目编号
        code (str): 生成的调用代码
    """
    url = urljoin(HOST, f"/api/v1/project/{xid}/code/update/")
    data = {"code": code}
    logger.debug(f"更新调用代码 {xid} {code}")
    resp = requests.post(url, json=data)
    logger.debug(resp.json())


def get_project_by_id(xid):
    """使用项目编号获取项目信息

    Args:
        xid (int): 项目编号
    """
    url = urljoin(HOST, f"/api/v1/project/{xid}/")
    resp = requests.get(url, timeout=30)
    if resp.status_code != 200:
        raise Exception("接口请求失败！")
    data = resp.json()
    if data["code"] != 0:
        raise Exception("接口请求失败！")
    logger.debug(resp.json())
    return data["data"]


def delete_project_by_id(xid):
    """使用项目编号获取项目信息

    Args:
        xid (int): 项目编号
    """
    url = urljoin(HOST, f"/api/v1/project/{xid}/delete/physical/")
    resp = requests.get(url, timeout=30)
    if resp.status_code != 200:
        raise Exception("接口请求失败！")
    data = resp.json()
    if data["code"] != 0:
        raise Exception("接口请求失败！")
    logger.debug(resp.json())
    return data["data"]


def load_model_class(model_name):
    pattern = "<class '(.*)'>"
    text = re.findall(pattern, model_name)[0]
    *model_path, model_name = text.split(".")
    model_path = ".".join(model_path)
    model_module = importlib.import_module(model_path)
    return getattr(model_module, model_name)



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/luwu/backend/v1.py
# @Desc         :
import json
import os
import re
import time

from flask import request
from flask.blueprints import Blueprint

from luwu.backend import status_code_wrapper
from luwu.backend.model import TrainProject, db
from luwu.core.models import image

api_v1_blueprint = Blueprint("api_v1_blueprint", __name__, url_prefix="/api/v1")

ENGINE_LIST = [
    {"index": 1, "name": "预设模型", "tip": "预先定义好的、结构固定的模型"},
    # {"index": 2, "name": "AutoKeras", "tip": ""},
    # {"index": 3, "name": "KerasTuner", "tip": ""},
    # {"index": 4, "name": "NNI", "tip": ""},
]


def check_path_correct(path):
    """检查给定路径是否符合要求，不符合要求则会抛出异常

    Args:
        path (str): 待检查路径
    """
    if not os.path.exists(path):
        raise Exception(f"指定路径 {path} 不存在！")
    if not os.path.isdir(path):
        raise Exception(f"{path} 必须是文件夹！")


@api_v1_blueprint.route("/")
def index():
    return "hello world"


@api_v1_blueprint.route("/image/classifier/engines/")
@status_code_wrapper()
def get_engine_list():
    data = ENGINE_LIST
    return data


@api_v1_blueprint.route("/image/classifier/models/<index>/")
@status_code_wrapper()
def get_image_classifier_list(index):
    index_engine_map = {item["index"]: item["name"] for item in ENGINE_LIST}
    engine = index_engine_map.get(int(index), "")
    if not engine:
        raise Exception("指定模型引擎不存在！")
    data = []
    if engine == "预设模型":
        classifiers = list(image.pre_trained_classifiers)
        data = []
        pattern = "Luwu(.*)ImageClassifier"
        for item in classifiers:
            cls_name = item.__name__
            model_name = re.findall(pattern, cls_name)[0]
            data.append({"index": cls_name, "name": model_name, "tip": ""})
    return data


@api_v1_blueprint.route("/image/classifier/project/create/", methods=["POST"])
@status_code_wrapper()
def create_image_classify_project():
    # 模型名称
    model_name = request.json.get("model_name", "")
    if not model_name:
        raise Exception("必须选择一个Model!")
    try:
        model = getattr(image, model_name)
    except:
        raise Exception("选择的模型不存在！")
    # 原始数据集
    dataset_index = request.json.get("dataset_index")
    origin_dataset_path = request.json.get("origin_dataset_path", "")
    check_path_correct(origin_dataset_path)
    # 清洗后的数据集保存路径
    target_dataset_path = request.json.get("target_dataset_path", "")
    if target_dataset_path:
        check_path_correct(target_dataset_path)
    else:
        target_dataset_path = origin_dataset_path
    # 模型保存路径
    model_save_path = request.json.get("model_save_path", "")
    if model_save_path:
        check_path_correct(model_save_path)
    else:
        model_save_path = target_dataset_path
    # 训练参数
    batch_size = int(request.json.get("batch_size", 32))
    epochs = int(request.json.get("epochs", 30))
    if dataset_index == 1:
        pass
    else:
        raise Exception("不支持的数据集类型！")
    train_project = TrainProject()
    train_project.params = {
        "dataset_index": dataset_index,
        "origin_dataset_path": origin_dataset_path,
        "target_dataset_path": target_dataset_path,
        "model_save_path": model_save_path,
        "batch_size": batch_size,
        "epochs": epochs,
    }
    train_project.model_name = str(model)
    train_project.status = 0
    train_project.addtime = int(time.time())
    train_project.add()


@api_v1_blueprint.route("/project/list/")
@status_code_wrapper()
def get_train_project_list():
    projects = db.session.query(TrainProject).all()
    data = [item.to_dict() for item in projects]
    return data


@api_v1_blueprint.route("/project/list/exists/")
@status_code_wrapper()
def get_train_project_exists_list():
    projects = db.session.query(TrainProject).filter(TrainProject.deleted == 0).all()
    data = [item.to_dict() for item in projects]
    return data


@api_v1_blueprint.route("/project/<xid>/")
@status_code_wrapper()
def get_train_project_by_id(xid):
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    data = tp.to_dict()
    return data


@api_v1_blueprint.route("/project/<xid>/delete/physical/")
@status_code_wrapper()
def physical_delete_train_project_by_id(xid):
    """物理删除指定训练项目"""
    db.session.query(TrainProject).filter(TrainProject.id == int(xid)).delete()
    db.session.commit()


@api_v1_blueprint.route("/project/<xid>/delete/logical/")
@status_code_wrapper()
def logical_delete_train_project_by_id(xid):
    """逻辑删除指定训练项目"""
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    tp.deleted = 1
    tp.status = 5
    db.session.commit()


@api_v1_blueprint.route("/project/<xid>/status/update/", methods=["POST"])
@status_code_wrapper()
def update_train_project_status(xid):
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    status = request.json.get("status", None)
    if status is None:
        raise Exception("指定状态不正确！")
    tp.status = status
    db.session.commit()


@api_v1_blueprint.route("/project/<xid>/code/update/", methods=["POST"])
@status_code_wrapper()
def update_train_project_code(xid):
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    code = request.json.get("code", None)
    if code is None:
        raise Exception("指定状态不正确！")
    tp.code = code
    db.session.commit()


@api_v1_blueprint.route("/project/<xid>/update/", methods=["POST"])
@status_code_wrapper()
def update_train_project_attr(xid):
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    for attr, value in request.json.items():
        setattr(tp, attr, value)
    db.session.commit()


@api_v1_blueprint.route("/project/<int:xid>/logs/", defaults={"start_line": 0})
@api_v1_blueprint.route("/project/<int:xid>/logs/<int:start_line>/")
@status_code_wrapper()
def read_project_logs_by_id(xid, start_line):
    tp = db.session.query(TrainProject).get(int(xid))
    if not tp:
        raise Exception("指定项目编号不存在！")
    model_save_path = tp.params["model_save_path"]
    log_path = os.path.join(
        model_save_path, f"luwu-classification-project-{xid}", f"train.log"
    )
    if not os.path.exists(log_path):
        raise Exception("暂无日志！请稍后再试...")
    lines = []
    with open(log_path, "r") as f:
        for index, line in enumerate(f):
            if index < start_line:
                continue
            if line.endswith("\n"):
                lines.append(line)
            else:
                break
    current_line = start_line + len(lines) - 1
    text = "<br/>".join(lines)
    if start_line > 0:
        text = "<br/>" + text
    data = {"current_line": current_line, "text": text}
    return data



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-11
# @FilePath     : /LuWu/bin/luwu.py
# @Desc         :
import argparse

parser = argparse.ArgumentParser(
    description="""
    Luwu命令行工具。
    你完全可以不写任何代码，就能够完成深度学习任务的开发。
    
    LuWu command tool.
    You can complete the development of deep learning tasks without writing any code.
    """
)
usage = """
usage: luwu_cmd_tool.py cmd

Luwu命令行工具。
你完全可以不写任何代码，就能够完成深度学习任务的开发。

LuWu command tool.
You can complete the development of deep learning tasks
without writing any code.

arguments:
    cmd: 命令参数，用以指定 Luwu 将要执行的操作
"""
# parser.add_argument(
#     "run web-server",
#     help="Run LuWu in the form of a web server.The default address is http://localhost:7788/ .",
#     type=str,
#     default="",
# )
subparsers = parser.add_subparsers(title="cmd", description="命令参数，用以指定 Luwu 将要执行的操作")

parse_web_server = subparsers.add_parser(
    "web-server", help="以web服务的形式启动Luwu，从而通过图形界面完成深度学习任务的开发。"
)
parse_web_server.set_defaults(cmd="web-server")

parse_object_detection = subparsers.add_parser("detection", help="通过命令行进行目标检测任务训练。")
parse_object_detection.set_defaults(cmd="detection")
parse_object_detection.add_argument(
    "--origin_dataset_path", help="处理前的数据集路径", type=str, default=""
)
parse_object_detection.add_argument(
    "--tfrecord_dataset_path", help="处理后的tfrecord数据集路径", type=str, default=""
)
parse_object_detection.add_argument(
    "--label_map_path", help="目标检测类别映射表(pbtxt)", type=str, default=""
)
parse_object_detection.add_argument(
    "--do_fine_tune", help="是否在预训练模型的基础上进行微调", type=bool, default=True
)
parse_object_detection.add_argument(
    "--fine_tune_checkpoint_path", help="预训练权重路径", type=str, default=""
)
parse_object_detection.add_argument(
    "--fine_tune_model_name",
    help='预训练模型名称。可选值的列表为["SSD ResNet50 V1 FPN 640x640 (RetinaNet50)","EfficientDet D0 512x512","CenterNet Resnet101 V1 FPN 512x512","CenterNet HourGlass104 512x512"]',
    type=str,
    default="",
)
parse_object_detection.add_argument(
    "--model_save_path", help="模型保存路径", type=str, default=""
)
parse_object_detection.add_argument(
    "--batch_size", help="mini batch 大小。默认 8.", type=int, default=8
)
parse_object_detection.add_argument(
    "--steps", help="训练steps数量. Defaults to 2000.", type=int, default=2000
)
parse_object_detection.add_argument(
    "--project_id", help="项目编号. Defaults to 0.", type=int, default=0
)

# TODO:增加分类任务的命令行执行
parse_classification = subparsers.add_parser("classification", help="通过命令行进行图像分类任务训练。")
parse_classification.set_defaults(cmd="classification")
parse_classification.add_argument(
    "network_name",
    help="分类器名称，支持的分类器有：[LuwuDenseNet121ImageClassifier,\
        LuwuDenseNet169ImageClassifier,LuwuDenseNet201ImageClassifier,\
        LuwuVGG16ImageClassifier,LuwuVGG19ImageClassifier,LuwuMobileNetImageClassifier,\
        LuwuMobileNetV2ImageClassifier,LuwuInceptionResNetV2ImageClassifier,\
        LuwuInceptionV3ImageClassifier,LuwuNASNetMobileImageClassifier,\
        LuwuNASNetLargeImageClassifier,LuwuResNet50ImageClassifier,\
        LuwuResNet50V2ImageClassifier,LuwuResNet101ImageClassifier,\
        LuwuResNet101V2ImageClassifier,LuwuResNet152ImageClassifier,\
        LuwuResNet152V2ImageClassifier,LuwuMobileNetV3SmallImageClassifier,\
        LuwuMobileNetV3LargeImageClassifier,LuwuXceptionImageClassifier,\
        LuwuEfficientNetB0ImageClassifier,LuwuEfficientNetB1ImageClassifier,\
        LuwuEfficientNetB2ImageClassifier,LuwuEfficientNetB3ImageClassifier,\
        LuwuEfficientNetB4ImageClassifier,LuwuEfficientNetB5ImageClassifier,\
        LuwuEfficientNetB6ImageClassifier,LuwuEfficientNetB7ImageClassifier]",
)
parse_classification.add_argument(
    "--origin_dataset_path", help="处理前的数据集路径", type=str, default=""
)
parse_classification.add_argument(
    "--target_dataset_path", help="处理后的tfrecord数据集路径", type=str, default=""
)
parse_classification.add_argument(
    "--model_save_path", help="模型保存路径", type=str, default=""
)
parse_classification.add_argument(
    "--validation_split", help="验证集切割比例。默认 0.2", type=float, default=0.2
)
parse_classification.add_argument(
    "--do_fine_tune", help="是进行fine tune，还是重新训练。默认 False", type=bool, default=False
)
parse_classification.add_argument(
    "--batch_size", help="mini batch 大小。默认 32.", type=int, default=8
)
parse_classification.add_argument(
    "--epochs", help="训练epoch数。默认 30.", type=int, default=30
)
parse_classification.add_argument(
    "--project_id", help="项目编号. Defaults to 0.", type=int, default=0
)

args = parser.parse_args()
print(args)


def run():
    if args.cmd == "web-server":
        from luwu import run

        run.run()
    elif args.cmd == "detection":
        from luwu.core.models.image import LuWuTFModelsObjectDetector

        LuWuTFModelsObjectDetector(**dict(args._get_kwargs())).run()
    elif args.cmd == "classification":
        class_name = args.network_name
        net_name = class_name.lstrip("Luwu").rstrip("ImageClassifier")
        import importlib

        luwu_image = importlib.import_module("luwu.core.models.image")
        # from luwu.core.models import image as luwu_image

        classifier_class = getattr(luwu_image, class_name)
        params = dict(args._get_kwargs())
        params["net_name"] = net_name
        classifier_class(**params).run()
    else:
        print("请检查指令是否有误！")


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-07
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /app/luwu/core/models/classifier/kerastuner/image_classifier.py
# @Desc         :
"""
这是还在规划中的功能，里面的代码目前都是无用的。
TODO:增加KerasTuner相关功能
"""
import tensorflow.keras as keras
from luwu.core.preprocess.image.load import (
    read_classify_dataset_from_dir,
    write_tfrecords_to_target_path,
)
from luwu.core.preprocess.image.process import (
    extract_image_and_label_from_record,
    normalized_image,
)
from tensorflow.keras import layers
import tensorflow as tf
from kerastuner.tuners.randomsearch import RandomSearch
import random
import os
from luwu.core.preprocess.data.data_generator import ImageClassifierDataGnenrator


class LuWuKerasTunerImageClassifier:
    def __init__(
        self,
        origin_dataset_path: str,
        target_dataset_path: str,
        max_trials: int = 20,
        executions_per_trial: int = 3,
        epochs: int = 10,
        batch_size: int = 4,
        validation_split: float = 0.2,
    ):
        self.origin_dataset_path = origin_dataset_path
        self.target_dataset_path = target_dataset_path
        self.batch_size = batch_size
        self.max_trials = max_trials
        self.executions_per_trial = executions_per_trial
        self.epochs = epochs
        self.validation_split = validation_split

    def _preprocess_dataset(self, dataset_path):
        dataset = tf.data.TFRecordDataset(dataset_path)
        dataset = dataset.map(extract_image_and_label_from_record)
        dataset = dataset.map(normalized_image)
        dataset = dataset.batch(self.batch_size)
        # dataset = dataset.prefetch(buffer_size=tf.data.experimental.AUTOTUNE)
        return dataset

    def preprocess_dataset(self):
        """对数据进行预处理"""
        # 读取原始数据
        data, classes_num_dict = read_classify_dataset_from_dir(
            self.origin_dataset_path
        )
        self.classes_num_dict = classes_num_dict
        # 切分数据
        total = len(data)
        dev_nums = int(total * self.validation_split)
        dev_data = random.sample(data, dev_nums)
        train_data = list(set(data) - set(dev_data))
        del data
        # 制作tfrecord数据集
        self.target_train_dataset_path = os.path.join(
            self.target_dataset_path, "train_dataset"
        )
        self.target_dev_dataset_path = os.path.join(
            self.target_dataset_path, "dev_dataset"
        )
        write_tfrecords_to_target_path(
            train_data, len(classes_num_dict), self.target_train_dataset_path
        )
        write_tfrecords_to_target_path(
            dev_data, len(classes_num_dict), self.target_dev_dataset_path
        )
        # 读取tfrecord数据集
        self.train_dataset = ImageClassifierDataGnenrator(
            self.target_train_dataset_path, batch_size=self.batch_size
        )
        self.dev_dataset = ImageClassifierDataGnenrator(
            self.target_dev_dataset_path, batch_size=self.batch_size, shuffle=False
        )

    def build_model(self, hp):
        model = keras.Sequential()
        model.add(layers.Input((224, 224, 3)))
        dropout_rate = hp.Float("dropout_rate", 0.1, 0.4, 0.05)
        conv_num_layers = hp.Int("conv_num_layers", 2, 3)
        dense_num_layers = hp.Int("dense_num_layers", 0, 1)
        for i in range(conv_num_layers):
            model.add(
                layers.Conv2D(
                    filters=hp.Int(f"filters_{i}", 8, 64, 8),
                    kernel_size=hp.Int(f"kernel_{i}", 3, 3, 2),
                    padding=hp.Choice(f"padding_{i}", ["same", "valid"]),
                    activation="relu",
                )
            )
            model.add(layers.Dropout(dropout_rate))
            if hp.Choice(f"maxpooling_{i}", [True, False]):
                model.add(layers.MaxPooling2D((2, 2), 2))
        model.add(layers.Flatten())
        for i in range(dense_num_layers):
            model.add(
                layers.Dense(units=hp.Int(f"units_{i}", 20, 60), activation="relu")
            )
        model.add(layers.Dense(len(self.classes_num_dict), activation="softmax"))
        model.compile(
            optimizer=keras.optimizers.Adam(
                hp.Choice("learning_rate", [1e-2, 1e-3, 1e-4])
            ),
            loss=keras.losses.categorical_crossentropy,
            metrics=["accuracy"],
        )
        return model

    def search(self):
        tuner = RandomSearch(
            self.build_model,
            objective="val_accuracy",
            max_trials=self.max_trials,
            executions_per_trial=self.executions_per_trial,
            directory="test_dir",
        )

        tuner.search_space_summary()

        tuner.search(
            self.train_dataset.for_fit(),
            epochs=10,
            steps_per_epoch=self.train_dataset.steps,
            validation_data=self.dev_dataset.for_fit(),
            validation_steps=self.dev_dataset.steps,
        )
        tuner.results_summary()
        print("--" * 20)
        print(tuner.get_best_hyperparameters()[0].get_config())

    def valid(self):
        model = keras.Sequential(
            [
                layers.Input((224, 224, 3)),
                layers.Conv2D(6, 3, activation="relu"),
                layers.Conv2D(16, 3, activation="relu"),
                layers.Flatten(),
                layers.Dense(3, activation="softmax"),
            ]
        )
        model.compile(
            optimizer=keras.optimizers.Adam(),
            loss=tf.keras.losses.categorical_crossentropy,
            metrics=["accuracy"],
        )
        model.summary()
        return model

    def run(self):
        # 预处理数据集
        self.preprocess_dataset()
        self.search()


if __name__ == "__main__":
    LuWuKerasTunerImageClassifier(
        "/Users/aaron/code/pets_classifer/images",
        "/Users/aaron/code/pets_classifer/images",
    ).run()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-20
# @Author       : AaronJny
# @LastEditTime : 2021-01-20
# @FilePath     : /LuWu/test/test_kerastuner_image_classifier.py
# @Desc         :
from luwu.core.models.classifier.kerastuner.image_classifier import LuWuKerasTunerImageClassifier

path = './images2'
luwu_classifier = LuWuKerasTunerImageClassifier(path, path)
luwu_classifier.run()



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /app/luwu/backend/config.py
# @Desc         :
import os


class Config:

    LUWU_DIR = os.path.expanduser("~/.luwu")

    SQLALCHEMY_DATABASE_URI = f"sqlite:///{LUWU_DIR}/luwu.db"

    PORT = 7788



# -*- coding: utf-8 -*-
# @Date         : 2021-01-20
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/luwu/core/preprocess/data/data_generator.py
# @Desc         :
import os

import tensorflow as tf
from jinja2 import Template
from luwu.core.preprocess.image.process import (
    extract_image_and_label_from_record,
    normalized_image,
    normalized_image_with_imagenet,
)


class BaseDataGenerator(object):
    def __init__(self, data_path, batch_size=32, image_size=224, shuffle=False):
        self.data_path = data_path
        self.batch_size = batch_size
        self.image_size = image_size
        self.shuffle = shuffle
        self._steps = -1
        self.dataset = self.load_dataset()

    def load_dataset(self):
        raise NotImplementedError

    def for_fit(self):
        yield from self.dataset

    @property
    def steps(self):
        if self._steps < 0:
            raise Exception("请在 `load_dataset` 中统计 steps!")
        else:
            return self._steps


class ImageClassifierDataGnenrator(BaseDataGenerator):
    def load_dataset(self):
        dataset = tf.data.TFRecordDataset(self.data_path)
        dataset = dataset.map(extract_image_and_label_from_record)
        dataset = dataset.map(lambda x, y: (x, y, self.image_size))
        dataset = dataset.map(normalized_image_with_imagenet)
        # todo:增加图像增广相关功能
        if self.shuffle:
            dataset = dataset.shuffle(10000)
        dataset = dataset.prefetch(self.batch_size).batch(self.batch_size)
        # 计算总步数
        cnt = 0
        for _ in dataset:
            cnt += 1
        self._steps = cnt
        dataset = dataset.repeat()
        return dataset

    def generate_preprocess_code(self):
        """生成数据处理的代码"""
        template_path = os.path.join(
            os.path.dirname(__file__), "templates/ImageClassifierDataGnenrator.txt"
        )
        with open(template_path, "r") as f:
            text = f.read()
        template = Template(text)
        return template.render(image_size=self.image_size)



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-10
# @FilePath     : /LuWu/luwu/core/models/complex/od/utils/generate_tfrecord.py
# @Desc         : 文件修改自 [https://tensorflow-object-detection-api-tutorial.readthedocs.io/en/latest/training.html](https://tensorflow-object-detection-api-tutorial.readthedocs.io/en/latest/training.html)
import argparse
import glob
import io
import os
import xml.etree.ElementTree as ET

import pandas as pd

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"  # Suppress TensorFlow logging (1)
from collections import namedtuple

import tensorflow.compat.v1 as tf
from object_detection.utils import dataset_util, label_map_util
from PIL import Image

# Initiate argument parser
parser = argparse.ArgumentParser(
    description="Sample TensorFlow XML-to-TFRecord converter"
)
parser.add_argument(
    "-x",
    "--xml_dir",
    help="Path to the folder where the input .xml files are stored.",
    type=str,
)
parser.add_argument(
    "-l", "--labels_path", help="Path to the labels (.pbtxt) file.", type=str
)
parser.add_argument(
    "-o", "--output_path", help="Path of output TFRecord (.record) file.", type=str
)
parser.add_argument(
    "-i",
    "--image_dir",
    help="Path to the folder where the input image files are stored. "
    "Defaults to the same directory as XML_DIR.",
    type=str,
    default=None,
)
parser.add_argument(
    "-c",
    "--csv_path",
    help="Path of output .csv file. If none provided, then no file will be " "written.",
    type=str,
    default=None,
)

args = parser.parse_args()

if args.image_dir is None:
    args.image_dir = args.xml_dir

label_map = label_map_util.load_labelmap(args.labels_path)
label_map_dict = label_map_util.get_label_map_dict(label_map)


def xml_to_csv(path):
    """Iterates through all .xml files (generated by labelImg) in a given directory and combines
    them in a single Pandas dataframe.

    Parameters:
    ----------
    path : str
        The path containing the .xml files
    Returns
    -------
    Pandas DataFrame
        The produced dataframe
    """

    xml_list = []
    for xml_file in glob.glob(path + "/*.xml"):
        tree = ET.parse(xml_file)
        root = tree.getroot()
        for member in root.findall("object"):
            value = (
                os.path.split(root.find("filename").text)[-1],
                int(root.find("size")[0].text),
                int(root.find("size")[1].text),
                member[0].text,
                int(member.find("bndbox").find("xmin").text),
                int(member.find("bndbox").find("ymin").text),
                int(member.find("bndbox").find("xmax").text),
                int(member.find("bndbox").find("ymax").text),
            )
            xml_list.append(value)
    column_name = [
        "filename",
        "width",
        "height",
        "class",
        "xmin",
        "ymin",
        "xmax",
        "ymax",
    ]
    xml_df = pd.DataFrame(xml_list, columns=column_name)
    return xml_df


def class_text_to_int(row_label):
    return label_map_dict[row_label]


def split(df, group):
    data = namedtuple("data", ["filename", "object"])
    gb = df.groupby(group)
    return [
        data(filename, gb.get_group(x))
        for filename, x in zip(gb.groups.keys(), gb.groups)
    ]


def create_tf_example(group, path):
    with tf.gfile.GFile(os.path.join(path, "{}".format(group.filename)), "rb") as fid:
        encoded_jpg = fid.read()
    encoded_jpg_io = io.BytesIO(encoded_jpg)
    image = Image.open(encoded_jpg_io)
    if image.mode != "RGB":
        rgb_image = image.convert("RGB")
        rgb_encode_jpg_io = io.BytesIO()
        rgb_image.save(rgb_encode_jpg_io, format="JPEG")
        encoded_jpg = rgb_encode_jpg_io.getvalue()

    width, height = image.size

    filename = group.filename.encode("utf8")
    image_format = b"jpg"
    xmins = []
    xmaxs = []
    ymins = []
    ymaxs = []
    classes_text = []
    classes = []

    for index, row in group.object.iterrows():
        xmins.append(row["xmin"] / width)
        xmaxs.append(row["xmax"] / width)
        ymins.append(row["ymin"] / height)
        ymaxs.append(row["ymax"] / height)
        classes_text.append(row["class"].encode("utf8"))
        classes.append(class_text_to_int(row["class"]))

    tf_example = tf.train.Example(
        features=tf.train.Features(
            feature={
                "image/height": dataset_util.int64_feature(height),
                "image/width": dataset_util.int64_feature(width),
                "image/filename": dataset_util.bytes_feature(filename),
                "image/source_id": dataset_util.bytes_feature(filename),
                "image/encoded": dataset_util.bytes_feature(encoded_jpg),
                "image/format": dataset_util.bytes_feature(image_format),
                "image/object/bbox/xmin": dataset_util.float_list_feature(xmins),
                "image/object/bbox/xmax": dataset_util.float_list_feature(xmaxs),
                "image/object/bbox/ymin": dataset_util.float_list_feature(ymins),
                "image/object/bbox/ymax": dataset_util.float_list_feature(ymaxs),
                "image/object/class/text": dataset_util.bytes_list_feature(
                    classes_text
                ),
                "image/object/class/label": dataset_util.int64_list_feature(classes),
            }
        )
    )
    return tf_example


def main(_):

    writer = tf.python_io.TFRecordWriter(args.output_path)
    path = os.path.join(args.image_dir)
    examples = xml_to_csv(args.xml_dir)
    grouped = split(examples, "filename")
    for group in grouped:
        tf_example = create_tf_example(group, path)
        writer.write(tf_example.SerializeToString())
    writer.close()
    print("Successfully created the TFRecord file: {}".format(args.output_path))
    if args.csv_path is not None:
        examples.to_csv(args.csv_path, index=None)
        print("Successfully created the CSV file: {}".format(args.csv_path))


if __name__ == "__main__":
    tf.app.run()



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-07
# @FilePath     : /LuWu/test/test_tf_object_detection_api.py
# @Desc         :
from luwu.core.models.complex.od.models import LuWuTFModelsObjectDetector


LuWuTFModelsObjectDetector(
    origin_dataset_path="/Users/aaron/od_api/VOCdevkit/VOC2007/yd_train_clean",
    tfrecord_dataset_path="/Users/aaron/test_luwu/dataset/",
    fine_tune_model_name="SSD ResNet50 V1 FPN 640x640 (RetinaNet50)",
    model_save_path="~/test_luwu/",
    batch_size=1,
).run()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-07
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /app/luwu/core/preprocess/image/load.py
# @Desc         :
import os
from glob import glob
from typing import List, Tuple

import tensorflow as tf
from tqdm import tqdm


def write_tfrecords_to_target_path(
    data: List[Tuple[str, str]], classes_num: int, target_dataset_path: str
):
    """将给定数据制作成tfrecords格式，并写入到目标路径

    Args:
        data (List[Tuple[str,str]]): 待写入数据
        classes_num (int): 共有多少类别
        target_dataset_path (str): 目标地址
    """
    with tf.io.TFRecordWriter(target_dataset_path) as writer:
        for image, num in tqdm(data):
            with open(image, "rb") as f:
                image_data = f.read()
            feature = {
                "image": tf.train.Feature(
                    bytes_list=tf.train.BytesList(
                        value=[
                            image_data,
                        ]
                    )
                ),
                "label": tf.train.Feature(int64_list=tf.train.Int64List(value=[num])),
                "num": tf.train.Feature(
                    int64_list=tf.train.Int64List(value=[classes_num])
                ),
            }
            example = tf.train.Example(features=tf.train.Features(feature=feature))
            writer.write(example.SerializeToString())


def read_classify_dataset_from_dir(dataset_path: str):
    """从给定地址读取并清洗按文件夹组织的图片数据集

    Args:
        dataset_path (str): 原始数据集地址
    """
    classes_num_dict = {}
    image_suffixes = {"jpg", "jpeg", "png"}
    data = []
    for sub_path in glob(os.path.join(dataset_path, "*")):
        if os.path.isdir(sub_path):
            class_name = sub_path.split("/")[-1]
            classes_num_dict[class_name] = len(classes_num_dict)
            for image in glob(os.path.join(sub_path, "*")):
                if image.split(".")[-1].lower() in image_suffixes:
                    data.append((image, classes_num_dict[class_name]))
    return data, classes_num_dict


if __name__ == "__main__":
    read_classify_dataset_from_dir(
        "/Users/aaron/code/pets_classifer/images",
        "/Users/aaron/code/pets_classifer/images/dataset",
    )



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/setup.py
# @Desc         :
import os
import sys

import setuptools
from setuptools import find_packages, setup
from setuptools.command.install import install
from setuptools.command.install_scripts import install_scripts

from install_script import run_install


class LuwuInstallCommand(install):
    def run(self):
        install.run(self)
        run_install()


class LuwuInstallScripts(install_scripts):
    def run(self):
        setuptools.command.install_scripts.install_scripts.run(self)

        # Rename some script files
        for script in self.get_outputs():
            if script.endswith(".py") or script.endswith(".sh"):
                dest = script[:-3]
            else:
                continue
            print("moving %s to %s" % (script, dest))
            with open(script, "r") as f:
                content = f.read()
            header = f"#!{sys.executable}\n"
            self.write_script(os.path.split(dest)[-1], header + content)
            os.remove(script)


setup(
    name="luwu",
    version="0.23",
    author="AaronJny",
    author_email="aaronjny7@gmail.com",
    description="LuWu——陆吾，一个简单的无代码深度学习平台。",
    url="https://github.com/AaronJny/luwu",
    packages=find_packages(include=["luwu.*", "luwu", "bin/*"]),
    include_package_data=True,
    python_requires=">=3.6.5",
    scripts=["bin/luwu.py"],
    cmdclass={
        "install": LuwuInstallCommand,
        "install_scripts": LuwuInstallScripts,
    },
)



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-01-28
# @FilePath     : /LuWu/luwu/backend/app.py
# @Desc         :
from flask import Flask
from flask_cors import CORS

from luwu.backend.config import Config
from luwu.backend.model import db


def create_app():
    app = Flask(
        __name__, static_folder="../dist", template_folder="../dist", static_url_path=""
    )
    app.config.from_object(Config)
    cors = CORS()

    cors.init_app(app)
    db.init_app(app)

    with app.app_context():
        # 初始化数据库
        db.create_all()

    from luwu.backend.v1 import api_v1_blueprint

    app.register_blueprint(api_v1_blueprint)

    from luwu.backend.views import main_blueprint

    app.register_blueprint(main_blueprint)

    return app


def run():
    app = create_app()
    app.run(host="0.0.0.0", port=Config.PORT)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-07
# @FilePath     : /LuWu/luwu/core/models/complex/od/utils/label_map_util.py
# @Desc         :
import glob
import os
import xml.etree.ElementTree as ET

from object_detection.utils.label_map_util import get_label_map_dict


def create_label_map(source_dataset_dir: str, pbtxt_save_path: str = "") -> str:
    """给定一个包含多个xml标注文件的目录，读取标注信息，生成label_map.pbtxt。

    Args:
        source_dataset_dir (str): xml所在的文件夹路径
        pbtxt_save_path (str, optional): label_map.pbtxt的保存路径，如果为空，则不保存. Defaults to ''.

    Returns:
        str: 生成的label_map.pbtxt的内容
    """
    label_dict = {}
    path = os.path.join(source_dataset_dir, "*.xml")
    for xml_file in glob.glob(path):
        tree = ET.parse(xml_file)
        root = tree.getroot()
        for member in root.findall("object"):
            label = member[0].text
            if label in label_dict:
                continue
            label_dict[label] = len(label_dict) + 1
    items = []
    item_text_format = """
    item {
        id: %s,
        name: '%s'
    }
    """
    for label, xid in label_dict.items():
        items.append(item_text_format % (xid, label))
    content = "\n".join(items)
    if pbtxt_save_path:
        dirname = os.path.dirname(pbtxt_save_path)
        os.makedirs(dirname, exist_ok=True)
        with open(pbtxt_save_path, "w") as f:
            f.write(content)
    return content



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-12
# @FilePath     : /LuWu/test/test_preset_image_classifier.py
# @Desc         :
from luwu.core.models.image import (
    LuwuResNet50V2ImageClassifier as ImageClassifier,
)

# path = "./images2/"
# path = "/app/images2"
ImageClassifier(
    origin_dataset_path="/Users/aaron/test_kaggle_api/chinese-click-demo-api/data",
    target_dataset_path="/Users/aaron/test_luwu",
    model_save_path="/Users/aaron/test_luwu",
    epochs=10,
).run()



# -*- coding: utf-8 -*-
# @Date         : 2021-01-07
# @Author       : AaronJny
# @LastEditTime : 2021-03-12
# @FilePath     : /LuWu/luwu/core/preprocess/image/process.py
# @Desc         :
import tensorflow as tf


def extract_image_and_label_from_record(example_string):
    feature_description = {  # 定义Feature结构，告诉解码器每个Feature的类型是什么
        "image": tf.io.FixedLenFeature([], tf.string),
        "label": tf.io.FixedLenFeature([], tf.int64),
        "num": tf.io.FixedLenFeature([], tf.int64),
    }
    feature_dict = tf.io.parse_single_example(example_string, feature_description)
    feature_dict["image"] = tf.io.decode_jpeg(feature_dict["image"], channels=3)
    # 将标签转成one-hot形式
    y = tf.cast(feature_dict["label"], dtype=tf.int32)
    y = tf.one_hot(y, tf.cast(feature_dict["num"], dtype=tf.int32))
    return feature_dict["image"], y


def normalized_image(image, label):
    # min_size = 28
    # max_size = 224
    # # 限定图片大小
    # image_size = max(min_size, image.shape[0])
    # image_size = min(image_size, max_size)
    image_size = 224
    # 缩放图片
    x = tf.image.resize(image, [image_size, image_size])
    # 将图片的像素值缩放到[0,1]之间
    x = tf.cast(x, dtype=tf.float32) / 255.0
    return x, label


def normalized_image_with_imagenet(image, label, image_size):
    # imagenet数据集均值
    image_mean = [0.485, 0.456, 0.406]
    # imagenet数据集标准差
    image_std = [0.299, 0.224, 0.225]
    # 缩放图片
    x = tf.image.resize(image, [image_size, image_size])
    # 将图片的像素值缩放到[0,1]之间
    x = tf.cast(x, dtype=tf.float32) / 255.0
    # 归一化
    x = (x - image_mean) / image_std
    return x, label



# -*- coding: utf-8 -*-
# @Date         : 2021-01-21
# @Author       : AaronJny
# @LastEditTime : 2021-03-12
# @FilePath     : /LuWu/luwu/core/models/classifier/preset/templates/LuwuPreTrainedImageClassifier.py
# @Desc         :
import tensorflow as tf


def read_image(path):
    return ""


# 定义编号->类别的映射
num_to_classes_map = {}
# 加载训练好的模型
model = tf.keras.models.load_model("")
# 加载并预处理图片
image = read_image("/path/to/image")
# 进行预测，并输出结果
outputs = model.predict(tf.reshape(image, (-1, 224, 224, 3)))
index = int(tf.argmax(outputs[0]))
print("当前图片类别为：{}".format(num_to_classes_map[index]))



# -*- coding: utf-8 -*-
# @Author       : AaronJny
# @LastEditTime : 2021-03-13
# @FilePath     : /LuWu/luwu/scripts/scheduler.py
# @Desc         :
import multiprocessing
import os
import subprocess
import time

from loguru import logger
from luwu.scripts.utils import (
    delete_project_by_id,
    read_train_projects,
    update_project_status,
)


def run_project(project):
    params = project["params"]
    model_save_path = params["model_save_path"]
    log_path = os.path.join(
        model_save_path,
        f"luwu-classification-project-{project['id']}",
        f"train.log",
    )
    curdir = os.path.abspath(os.path.dirname(__file__))
    cd_path = os.path.abspath(os.path.join(os.path.join(curdir, ".."), ".."))
    py_path = os.path.join("." + curdir[len(cd_path) :], "train_project.py")
    cmd = f"""cd {cd_path};python {py_path} {project['id']} > {log_path} 2>&1"""
    st, out = subprocess.getstatusoutput(cmd)
    if st == 0:
        logger.info("处理成功！")
    else:
        logger.info("处理失败！")
        raise Exception(out)


def run():
    # 项目编号->进程句柄的映射
    xid_process_map = {}
    while True:
        projects = read_train_projects()
        cnt = 0
        for xid, p in xid_process_map.items():
            if p.is_alive():
                logger.debug(f"项目 {xid}, 进程 {p.pid}, 活跃中...")
                cnt += 1
        logger.info(f"当前活跃进程数 {cnt}!")
        # logger.info(projects)
        # 按照返回的信息，纠正每个任务的状态
        for project in projects:
            xid = project["id"]
            status = project["status"]
            # 检查不该处于运行状态的任务，是否有运行的。
            # 如果有，就终止它
            if status not in (1, 2):
                # logger.info(f"!1 {project}")
                p: multiprocessing.Process = xid_process_map.get(xid, None)
                if p and p.is_alive():
                    logger.info(
                        f"发现一个异常运行的任务 project {xid} | process {p.pid}，尝试 terminate..."
                    )
                    p.terminate()
                if p and p.is_alive():
                    logger.info(
                        f"发现一个异常运行的任务 project {xid} | process {p.pid}，尝试 kill..."
                    )
                    p.kill()
                if xid in xid_process_map:
                    del xid_process_map[xid]
            # 检查是否存在待处理任务
            # 如果有，并且没有处于运行状态，就运行它
            if status == 1:
                logger.info(f"发现一个待处理任务 project {xid}...")
                p: multiprocessing.Process = xid_process_map.get(xid, None)
                if p:
                    if p.is_alive():
                        logger.info(f"任务已经存在 project {xid} | process {p.pid}！跳过")
                    # 进程正常退出
                    elif p.exitcode == 0:
                        update_project_status(xid, 3)
                        del xid_process_map[xid]
                    # 进程异常退出
                    else:
                        update_project_status(xid, 6)
                        del xid_process_map[xid]
                else:
                    logger.info(f"创建一个新的任务 project {xid}...")
                    p = multiprocessing.Process(target=run_project, args=(project,))
                    p.start()
                    xid_process_map[xid] = p
                    logger.info(f"创建完成！project {xid} | process {p.pid}")
            # 检查处于运行中状态的任务
            # 如果任务已经结束了，就更新状态
            if status == 2:
                p: multiprocessing.Process = xid_process_map.get(xid, None)
                if p:
                    # 进程存活
                    if p.is_alive():
                        continue
                    # 进程正常退出
                    elif p.exitcode == 0:
                        update_project_status(xid, 3)
                        del xid_process_map[xid]
                    # 进程异常退出
                    else:
                        update_project_status(xid, 6)
                        del xid_process_map[xid]
                else:
                    update_project_status(xid, 6)
        # 对于已经停止的、且处理逻辑删除状态的任务，执行物理删除
        for project in projects:
            xid = project["id"]
            deleted = project["deleted"]
            if deleted and xid not in xid_process_map:
                delete_project_by_id(xid)
        time.sleep(10)


if __name__ == "__main__":
    run()
